@@ -1,9 +1,10 @@
 # frozen_string_literal: true
+
 module Bootsnap
   module ExplicitRequire
-    ARCHDIR    = RbConfig::CONFIG['archdir']
-    RUBYLIBDIR = RbConfig::CONFIG['rubylibdir']
-    DLEXT      = RbConfig::CONFIG['DLEXT']
+    ARCHDIR    = RbConfig::CONFIG["archdir"]
+    RUBYLIBDIR = RbConfig::CONFIG["rubylibdir"]
+    DLEXT      = RbConfig::CONFIG["DLEXT"]
 
     def self.from_self(feature)
       require_relative("../#{feature}")
