@@ -1,17 +1,21 @@
 # frozen_string_literal: true
 
-require("test_helper")
-require("bootsnap/load_path_cache")
+require "test_helper"
+require "bootsnap/load_path_cache"
 
 module Bootsnap
   module LoadPathCache
-    class PathTest < MiniTest::Test
+    class PathTest < Minitest::Test
+      include LoadPathCacheHelper
+
       def setup
+        super
         @cache = Object.new
       end
 
       def test_stability
-        require("time")
+        require "time"
+
         time_file       = Time.method(:rfc2822).source_location[0]
         volatile        = Path.new(__FILE__)
         stable          = Path.new(time_file)
